/*
 * Thanks to Scot Laird, Russell Nelson, and others..
 *
 * Copyright (C) 1996 by Thomas Davis
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL SCOTT LAIRD BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#define __KERNEL__
#define MODULE

#include <linux/module.h>

#include <linux/types.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/config.h>
#include <linux/malloc.h>
#include <linux/delay.h>

#include <asm/io.h>

#include "quickcam.h"
#include "qcam-kernel.h"

int quickcam_major = 0;

struct quickcam camera, *qc;

int lp_ports[] = { 0x3BC, 0x278, 0x378 };

/*
 * One of the two ways to generate a scan.. seek to the beginning of the
 * frame buffer via offset 0, orig 0.
 *
 * We also disallow any seeks past the end of the frame buffer,
 * and any seeks before the beginning of the buffer..
 */

static int qcam_lseek(struct inode *inode, struct file *file, off_t ofs, int origin)
{
	if (qc->debug > 1) {
		printk("\nquickcam: lseek ofs=%ld, origin=%d\n", ofs, origin);
	}
 
	switch (origin) {
		case 0:
			file->f_pos = ofs;
			if (ofs == 0) {
				if (doscan(qc) < 0) {
					return -ENODEV;
				}
			}
			break;
		case 1:
			file->f_pos += ofs;
			break;
		case 2:
			/* SEEK_END */
		default:   
			return  -EINVAL;
	}
 
	if ((file->f_pos < 0) || ( file->f_pos > FRAME_BUFFER_SIZE)) {
		file->f_pos = 0;
		return -ESPIPE;
	}

	if (qc->debug > 1)
		printk("quickcam origin set to %lu\n", file->f_pos);
   
	return file->f_pos;
}

/*
 * This is tricky.  To get a scan, you must do a lseek(fd, 0, FD_SET).
 * We don't scan here, just in cause there is another program wanting to
 * just attach to the device to send ioctl's.
 *
 */

static int qcam_open(struct inode *inode, struct file *file)
{
	if (qc->debug > 1) {
		printk("\nquickcam open: dev minor=%d\n", MINOR(inode->i_rdev));
		printk("quickcam open: dev major=%d\n", MAJOR(inode->i_rdev));
		printk("reseting camera...\n");
	}

	if (qc->buffer == NULL) {
		qc->buffer = (unsigned char *) vmalloc(FRAME_BUFFER_SIZE);
	}

	qc->init = 1;
	qc->framecount = 0;

	reset_quickcam(qc);

	if (qc->debug > 1) {
		printk("initing camera...\n");
	}

#if 0
	qcam_init(qc);

	if (qc->debug > 1) {
		printk("scanning in %d bpp mode...\n", qc->bpp);
	}

	if (doscan(qc) < 0) {
		return -ENODEV;
	}

#endif
	if (qc->debug > 1)
		printk("quickcam open done\n" );

        MOD_INC_USE_COUNT;

	return (0);
}

static void qcam_close(struct inode *inode, struct file *file)
{
	if (qc->debug > 1)
		printk("quickcam close OK\n" );

#if 0
	if (qc->buffer != NULL) 
		vfree(qc->buffer);
#endif
        MOD_DEC_USE_COUNT;
}

static int qcam_read(struct inode *inode, struct file *file, char *buf, int count)
{
	int read = 0, chars = 1, size = 0;
   
	if (qc->debug > 1)
		printk("\nquickcam: read request of %d bytes\n",count);
   
	if (file->f_pos > FRAME_BUFFER_SIZE)
		return -EIO;

	if (qc->debug > 4) 
		printk("quickcam: before while read = %d, chars = %d, size = %d, count = %d \n", read, chars, size, count); 

	while (chars > 0 && count > 0 && file->f_pos <= FRAME_BUFFER_SIZE) {
		chars = FRAME_BUFFER_SIZE - file->f_pos;
		if (chars > count)
			chars = count;
		if (chars > FRAME_BUFFER_SIZE)
			chars = FRAME_BUFFER_SIZE;
		memcpy_tofs(buf, (qc->buffer) + file->f_pos, chars);
		read += chars;
		file->f_pos += chars;
		count -= chars;
		if (qc->debug > 4) 
			printk("quickcam: while - read = %d, chars = %d, size = %d, count = %d \n", read, chars, size, count); 
	}

	if (qc->debug > 1)
		printk("quickcam: read = %d \n", read); 

	return read;
}

static int qcam_write(struct inode *inode, struct file *filp, const char *ostr, int count)
{
	if (qc->debug > 1)
		printk("\nquickcam: try to write %d bytes\n", count);

	return (0);
}

static int qcam_ioctl(struct inode *inode, struct file * filp, unsigned int cmd, unsigned long arg)
{
        int retval = 0;
        struct qcam camera;

	if (qc->debug > 1)
		printk("quickcam: trying ioctl %d\n",cmd);

	switch (cmd) {
		case QC_GET:
			retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(camera));

			if (retval)
				return retval;

			camera.xsize = qc->xsize;
			camera.ysize = qc->ysize;
			camera.bpp = qc->bpp;
			camera.mode = qc->mode;
			camera.brightness = qc->brightness;
			camera.contrast = qc->contrast;
			camera.whitebalance = qc->whitebalance;
			camera.status = qc->status;
			camera.mode = qc->mode;
			camera.zoom = qc->zoom;
			camera.debug = qc->debug;
			camera.framecount = qc->framecount;
			memcpy_tofs((int *) arg, &camera, sizeof(camera));
			break;

		case QC_SET:
                        retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(camera));
			
                        if (retval)
                                return retval;

                        memcpy_fromfs(&camera, (void *) arg, sizeof(camera));
			qc->xsize = camera.xsize;
			qc->ysize = camera.ysize;
			qc->bpp = camera.bpp;
			qc->mode = camera.mode;
			qc->brightness = camera.brightness;
			qc->contrast = camera.contrast;
			qc->whitebalance = camera.whitebalance;
			qc->status = camera.status;
			qc->debug = camera.debug;
			qc->init = 1;
			qc->framecount = 0;
                        break;

		case QC_SET_BRIGHTNESS:
                        retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
			
                        if (retval)
                                return retval;

                        memcpy_fromfs(&qc->brightness, (void *) arg, sizeof(int));
                        break;

		case QC_GET_BRIGHTNESS:
			retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));

			if (retval)
				return retval;

			memcpy_tofs((int *) arg, &qc->brightness, sizeof(int));
			break;

		case QC_SET_CONTRAST:
                        retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
			
                        if (retval)
                                return retval;

                        memcpy_fromfs(&qc->contrast, (void *) arg, sizeof(int));
			qc->init = 1;
                        break;

		case QC_GET_CONTRAST:
			retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));

			if (retval)
				return retval;

			memcpy_tofs((int *) arg, &qc->contrast, sizeof(int));
			break;

		case QC_SET_WHITE_BALANCE:
                        retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
			
                        if (retval)
                                return retval;

                        memcpy_fromfs(&qc->whitebalance, (void *) arg, sizeof(int));
			qc->init = 1;
                        break;

		case QC_GET_WHITE_BALANCE:
			retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));

			if (retval)
				return retval;

			memcpy_tofs((int *) arg, &qc->whitebalance, sizeof(int));
			break;

		default:
			retval = -EINVAL;
	}
	return retval;
}

static struct file_operations fops = {
   qcam_lseek,			/* lseek   */
   qcam_read,			/* read    */
   qcam_write,			/* write   */
   NULL,                        /* readdir */  
   NULL,                        /* select  */ 
   qcam_ioctl,			/* ioctl   */
   NULL,                        /* mmap    */
   qcam_open,			/* open    */
   qcam_close			/* close   */
};

int init_module(void) 
{
	int ret;
	int i;

	qc = (struct quickcam *) &camera;

	qc->port = 0;
	qc->debug = 1;

	printk("quickcam: init module \n");

	if (qc->debug > 3) {
		printk("quickcam: port=0x%x debug=%d\n", qc->port, qc->debug);
	}
   
	ret = register_chrdev (quickcam_major, "quickcam", &fops);

	if (ret < 0) {
		printk("quickcam: register device failed ret = %d\n", ret);
		goto error_module;
	}

	if (quickcam_major == 0)
		quickcam_major = ret;
   
	for (i = 0; i < 3; i++) {
		printk("quickcam: trying port 0x%x ", lp_ports[i]);
		if ((ret = check_region(lp_ports[i], 3)) ) {
			printk("which is already in use by the system..\n");
			continue;
		}

		ret = qcam_find(lp_ports[i]);

		if (!ret) {
			printk("which has a camera attached to it.\n");
			qc->port = lp_ports[i];
			break;
		}
	}

	if (qc->port == 0) {
		printk("quickcam: No camera found.  Please check cables or system..\n");
		ret = 0;
		goto error_device;
	}

	if ((ret = check_region(qc->port, 3)) ) {
		printk("quickcam: port %x in use.  Please pick another port.\n", qc->port);
		ret = 0; 
		goto error_device;
	}

	reset_quickcam(qc);

	if (qc->status != 0x75) {
		printk("quickcam: port 0x%x is bi-directional! Camera will be on fire!\n", qc->port);
	}

	request_region(qc->port, 3, "quickcam");

	printk("quickcam: device driver loaded\n");

	qc->contrast = 100;
	qc->brightness = 125;
	qc->whitebalance = 100;
	qc->xfermode = 0x07;
	qc->xsize = 160;
	qc->ysize = 120;
	qc->init = 1;
	qc->panx = 0x01;
	qc->pany = 0x07;
	qc->buffer = NULL;

	return 0;

#if 0
	error_ports:
		release_region(qc->port,8);
#endif

	error_device:
		unregister_chrdev(quickcam_major,"quickcam");

	error_module:
		printk("quickcam: device driver load failed\n");

	return (ret)?ret:-EINVAL;
}


void cleanup_module (void) 
{
	if (qc->debug > 1)
		printk("quickcam: cleanup module\n");

	if (qc->buffer != NULL) 
		vfree(qc->buffer);

	release_region(qc->port, 3);
   
	unregister_chrdev(quickcam_major, "quickcam");
}
